/*
 * Scheme modules
 */
#include "s.h"
#include "vm2.h"
#include "stack.h"

SOBJ scm_global_module;			/* the global module */
SOBJ scm_current_module;		/* the current module */

SOBJ scm_module_hash;			/* contains ref to all modules */


/****************************************************************
 * Type driver functions 
 ****************************************************************/
void scm_module_mark(SOBJ obj)
{
  SCM_Module *m = SCM_MODULE(obj);
  if (m->name) { 		scm_gc_mark(m->name); }
  if (m->symbols) { 	scm_gc_mark(m->symbols); }
  if (m->exports) { 	scm_gc_mark(m->exports); }
  if (m->imports) { 	scm_gc_mark(m->imports); }
}

void scm_module_sweep(SOBJ obj)
{
  if (SCM_MODULE(obj)) {  scm_free(SCM_MODULE(obj)); }
  SCM_MODULE(obj) = NULL;
}


void scm_module_print(SOBJ obj, PORT *p)
{
  port_puts(p, "<#module ");
  scm_write_obj(SCM_MODULE(obj)->name, p, TRUE);
  port_puts(p, ">");
}

void scm_module_write(SOBJ obj, PORT *p)
{
  scm_module_print(obj,p);
}

/****************************************************************
 * helper functions
 ****************************************************************/

/* Search symbol in a module.
 *
 * Only search in the specified module (0 depth search). Returns the
 * symbol entry or null if not found */

SOBJ scm_module_symbol_search(SOBJ mod, SOBJ atom)
{
  SCM_Module *m;
  SOBJ sym;

  if (!SCM_MODULEP(mod)) 	err("module-search: bad module", mod);
  if (!SCM_ATOMP(atom)) 	err("module-search: bad key", atom);
  
  m = SCM_MODULE(mod);
  sym = scm_hash_search(m->symbols, atom);
  return( sym == scm_undefined ? NULL : sym );
}

/* Search symbol in a module.
 *
 * If not found, create a new symbol with unbound value
 */
SOBJ scm_module_symbol_lookup(SOBJ mod, SOBJ atom)
{
  SCM_Module *m;
  SOBJ sym;

  if ( (sym = scm_module_symbol_search(mod, atom)) != NULL) 
	return(sym);

  m = SCM_MODULE(mod);
  return(SCM_CAR(scm_hash_set(m->symbols, atom, scm_unbound)));
}

/* find symbol in export list
 *
 * Try to find symbol in export list of imported modules of the
 * module. Returns a pointer to the symbol or NULL if search failed.
 */

SOBJ scm_module_find_exported_symbol(SCM_Module *m, SOBJ atom)
{
  SOBJ list, sym, module, obj;

  /*-- scan each imported module */

  for (list = m->imports; list != NULL; list = SCM_CDR(list)) {
	module = SCM_CAR(list);

	/*-- search exported symbol list */

	for (sym = SCM_MODULE(module)->exports; sym != NULL; sym = SCM_CDR(sym)) {

	  if (SCM_SYM_NAME(SCM_CAR(sym)) == atom) {
		obj = scm_module_symbol_search(module, atom);
		if (obj != NULL) return(obj);
		
		scm_puts("OOPS: symbol "); scm_cdisplay(atom);
		scm_puts(" is defined in export list of module ");
		scm_cdisplay(module);
		scm_puts(" but cannot be found in symbol table\n");
		break;
	  }
	}
	/*-- not found in the export list of this module. Try parent module */
	
	obj = scm_module_find_exported_symbol(SCM_MODULE(module), atom);
	if (obj != NULL) 	return(obj);
	
  }
  return(NULL);
}

/* Find symbol in module.
 *
 * First search current module, then search in imported module and
 * finally search in the global module.
 *
 * If create_it is true, we create the symbol if it is not found in
 * current and in explicitly imported modules.
 */
SOBJ scm_module_find_symbol(SOBJ mod, SOBJ atom, int create_it)
{
  SCM_Module *m;
  SOBJ sym;
  
  m = SCM_MODULE(mod);

  /*-- search in the specified */
  if ((sym = scm_module_symbol_search(mod, atom)) != NULL)
	return(sym);


  /*-- search in imported module chain */
  sym = scm_module_find_exported_symbol(m, atom);
  if (sym != NULL)
	return(sym);

  /*-- last chance: search in the global module */
  if ((sym = scm_hash_search(scm_symbol_hash, atom)) != scm_undefined)
	return(sym);
  
  if (create_it) {
	sym = SCM_CAR(scm_hash_set(m->symbols, atom, scm_unbound));
	return(sym);
  }	

  return(NULL);
}

/****************************************************************
 * Scheme functions for modules
 ****************************************************************/

SOBJ scm_make_module(SOBJ name)
{
  SCM_Module *m;
  SOBJ new;

  if (!SCM_ATOMP(name)) err("make-module: bad atom", name);

  new = scm_hash_ref(scm_module_hash, name);

  if (new == scm_undefined) {

	/* module does not exist. have to create it */
	new = scm_newcell(SOBJ_T_MODULE);
	m = scm_must_alloc(sizeof(SCM_Module));
	
	SCM_MODULE(new) = m;
	m->name = name;
	m->symbols = scm_mkhash(SCM_HASH_T_SYMBOL); 	/* hash for symbols */
	m->imports = NULL;				/* list of imported module */
	m->exports = NULL;				/* list of exported symbols */
	m->export_all = FALSE;			/* default to export list */
	scm_hash_set(scm_module_hash, name, new);
  }
  return(new);
}


SOBJ scm_modulep(SOBJ obj)
{
  return(SCM_MKBOOL(SCM_MODULEP(obj)));
}

SOBJ scm_get_current_module()
{
  return(scm_current_module);
}

SOBJ scm_set_current_module(SOBJ mod)
{
  if (SCM_STRINGP(mod)) {
	mod = scm_hash_ref(scm_module_hash, mod);
  }
  if (SCM_SYMBOLP(mod)) {
	mod = scm_hash_ref(scm_module_hash, scm_symbol_to_string(mod));
  }

  if (!SCM_MODULEP(mod))	err("set-current-module: bad module", mod);
  scm_current_module = mod;
  return(mod);
}

SOBJ scm_import(int argc, SOBJ *argv)
{
  SCM_Module *m = SCM_MODULE(scm_current_module);
  SOBJ *l = argv+argc;
  SOBJ *p = argv;
  SOBJ n;

  while(p < l) {
	if (!SCM_MODULE(*p)) 	err("import: bad module", *p);
	n = m->imports;
	if (n == NULL) {
	  m->imports = scm_cons(*p, NULL);
	} else {
	  while(n != NULL) {
		if (SCM_CAR(n) == *p)	break;
		if (SCM_CDR(n) == NULL) {
		  SCM_CDR(n) = scm_cons(*p, NULL);
		  break;
		}
		n = SCM_CDR(n);
	  }
	}
	p++;
  }
  return(m->imports);
}

SOBJ scm_export(int argc, SOBJ *argv)
{
  SCM_Module *m = SCM_MODULE(scm_current_module);

  return(m->exports);
}

SOBJ scm_module_exports(SOBJ mod)
{
  if (!SCM_MODULEP(mod)) err("module-exports: bad module", mod);
  return(SCM_MODULE(mod)->exports);
}

SOBJ scm_module_imports(SOBJ mod)
{
  if (!SCM_MODULEP(mod)) err("module-imports: bad module", mod);
  return(SCM_MODULE(mod)->imports);
}

SOBJ scm_module_name(SOBJ mod)
{
  if (!SCM_MODULEP(mod)) err("module-name: bad module", mod);
  return(SCM_MODULE(mod)->name);
}

SOBJ scm_module_symbols(SOBJ mod)
{
  if (!SCM_MODULEP(mod)) err("module-symbols: bad module", mod);
  return(SCM_MODULE(mod)->symbols);
}

SOBJ scm_find_module(SOBJ name)
{
  if (SCM_ATOMP(name)) {
	return(scm_hash_ref(scm_module_hash, name));
  }
  if (SCM_STRINGP(name)) {
	return(scm_hash_ref(scm_module_hash, scm_mkatom(SCM_STR_VALUE(name))));
  }
  return(scm_false);
}

/*-- (define-module <name> <body>) */

static void refresh_mod_symbols(SOBJ expr)
{
  SOBJ obj;

  while(expr) {
	obj = SCM_CAR(expr);
	if (SCM_PAIRP(obj)) { refresh_mod_symbols(obj); }
	if (SCM_SYMBOLP(obj)) {
	  SCM_CAR(obj) = scm_sym_clone(SCM_CAR(obj));
	}
	expr = SCM_CDR(expr);
  }
}


SOBJ scm_syntax_module(SOBJ icode, SOBJ expr, SOBJ env)
{
  SOBJ name;
  SOBJ old_mod, new_mod;

  name = SCM_CAR(expr);
  if (!SCM_ATOMP(name))		err("define-module: bad module name", name);
  expr = SCM_CDR(expr);
  
  scm_puts("module: name=");
  scm_cdisplay(name);	scm_puts(" expr=");		scm_cprint(expr);

  old_mod = scm_current_module;
  new_mod = scm_make_module(name);
  SCM_SYM_VALUE(name) = new_mod;

  scm_current_module = new_mod;

  /*-- we have to resymbol all the expr before compiling it. */
  /* refresh_mod_symbols(expr); */

  scm_compile_begin(icode, expr, env);
  scm_current_module = old_mod;
  return(icode);
}

SOBJ scm_syntax_export(SOBJ expr, SOBJ env)
{
  SCM_Module *m;
  SOBJ atom, p;
  int found;

  scm_puts("; export: expr="); scm_cprint(expr);
  scm_puts(" env="); scm_cprint(env);

  m = SCM_MODULE(scm_current_module);
  while(expr) {
	atom = SCM_CAR(expr);
	if (!SCM_ATOMP(atom)) err("export: bad atom", atom);

	found = FALSE;
	for (p = m->exports; p ; p = SCM_CDR(p)) {
	  if (SCM_CAR(p) == atom) found = TRUE;
	}
	if (!found) {
	  m->exports = scm_cons(atom, m->exports);
	  scm_module_symbol_lookup(scm_current_module, atom);
	}	  
	expr = SCM_CDR(expr);
  }
  spush(m->exports);
  spush_opc(SCM_OP_PUSH);
  return(NULL);
}

SOBJ scm_syntax_import(SOBJ expr, SOBJ env)
{
  scm_puts("; import: expr="); scm_cprint(expr);
  scm_puts(" env="); scm_cprint(env);

  return(NULL);
}

/*-- reconizer */
int scm_module_wreconize(PORT *port, char *s)
{
  return( strstr(s, "::") != NULL );
}

SOBJ scm_module_wparse(PORT *port, char *str)
{
  SOBJ mod, sym;
  char module[128], symbol[128];
  char *p = strstr(str, "::");
  if (p != NULL) {
	strncpy(module, str, p - str);
	module[p - str] = 0;
	strcpy(symbol, p+2);
	mod = scm_find_module(scm_mkatom(module));
	sym = scm_module_find_symbol(mod, scm_mkatom(symbol), FALSE);
	if (sym) return(sym);
  }
  return(scm_undefined);
}


/****************************************************************
 * Init module system
 ****************************************************************/
void scm_init_module()
{
  scm_puts("; init modules...\n");

  scm_module_hash = scm_mkhash(SCM_HASH_T_SYMBOL);
  scm_gc_protect(&scm_module_hash);

  /*-- create the global module */
  scm_global_module = scm_make_module(scm_mkatom("global"));
  SCM_MODULE(scm_global_module)->symbols = scm_symbol_hash;
  SCM_MODULE(scm_global_module)->export_all = TRUE;
  
  /*-- set the current module to the global one */
  scm_current_module = scm_global_module;
  
  scm_add_cprim("make-module", 		scm_make_module,	1);
  scm_add_cprim("module?",			scm_modulep,		1);

  scm_add_cprim("current-module",	scm_get_current_module, 	0);
  scm_add_cprim("set-current-module", scm_set_current_module, 	1);
  scm_add_cprim("import",			scm_import,					-1);
  
  scm_add_cprim("module-exports",	scm_module_exports,			1);
  scm_add_cprim("module-imports",	scm_module_imports,			1);
  scm_add_cprim("module-symbols",	scm_module_symbols,			1);
  scm_add_cprim("module-name",		scm_module_name,			1);
  scm_add_cprim("find-module",		scm_find_module,			1);

  scm_add_syntax("module",			scm_syntax_module);
  scm_add_syntax("export",			scm_syntax_export);
  scm_add_syntax("import",			scm_syntax_import);
  scm_add_cvar("module-hash",		&scm_module_hash);
}
