/*
 * File and string ports
 *
 * $Id: port.c 1.19 Wed, 28 Apr 1999 07:03:23 +0200 crad $
 *
 */

#include "s.h"
#include "vm2.h"
#include "heap.h"
#include <unistd.h>

/************************************************************************
 * Variables
 ************************************************************************/
PORT *scm_inp, *scm_outp, *scm_errp;

SOBJ scm_in_port;
SOBJ scm_out_port;
SOBJ scm_err_port;
SOBJ scm_eof;

#define SCM_INP		SCM_PORT(scm_in_port);
#define SCM_OUTP	SCM_PORT(scm_out_port);
#define SCM_ERRP	SCM_PORT(scm_err_port);

/************************************************************************
 * Common port routines
 ************************************************************************/

static PORT *port_new(int type, int flag)
{
  PORT *p = scm_must_alloc(sizeof(PORT));
  scm_mem_clear(p, sizeof(PORT));
  p->type = type;
  p->flag = flag;
  return(p);
}


/************************************************************************
 * File ports
 ************************************************************************/

static int port_file_getc(PORT *p)
{
  int c;
  if ((c = getc(p->descr.f)) == EOF)  return(PORT_EOF);
  if (c == '\n') p->line++;
  return(c);
}

static int port_file_peekc(PORT *p)
{
  int c;
  if ((c = getc(p->descr.f)) == EOF)  return(PORT_EOF);
  ungetc(c, p->descr.f);
  return(c);
}

static void port_file_putc(PORT *p, char c)
{
  putc(c, p->descr.f);
  if (c == '\n') p->line++;
}

static void port_file_close(PORT *p)
{
  fclose(p->descr.f);
  scm_free(p);
}

static PORT *port_file_open(void *filename, int mode)
{
  PORT *port;
  FILE *fd;
  char sm[5];

  if (mode == PORT_RW) {
	strcpy(sm, "r+");
  } else if (mode & PORT_R) {
	strcpy(sm, "r");
  } else {
	strcpy(sm, "w");
  }
  fd = fopen(filename, sm);
  if (fd) {
	port = port_new(PORT_T_FILE, mode);
	port->descr.f = fd;

	return(port);
  }
  return(NULL);
}

/************************************************************************
 * string ports
 ************************************************************************/

static int port_string_getc(PORT *p)
{
  if (p->descr.s.index >= p->descr.s.length)	return(PORT_EOF);
  return(p->descr.s.data[p->descr.s.index++]);
}

static int port_string_peekc(PORT *p)
{
  if (p->descr.s.index >= p->descr.s.length)	return(PORT_EOF);
  return(p->descr.s.data[p->descr.s.index]);
}

static void port_string_putc(PORT *p, char c)
{
  if (p->descr.s.index >= (p->descr.s.alloced-1)) {
	if (p->descr.s.data != NULL) {
	  p->descr.s.data = scm_must_realloc(p->descr.s.data,
										 p->descr.s.alloced + PORT_STR_QTUM);
	  p->descr.s.alloced += PORT_STR_QTUM;
	} else {
	  p->descr.s.data = scm_must_alloc(PORT_STR_QTUM);
	  p->descr.s.alloced = PORT_STR_QTUM;
	}
  }
  p->descr.s.data[p->descr.s.index++] = c;
  p->descr.s.length = p->descr.s.index;
  p->descr.s.data[p->descr.s.index] = 0;
}

char *port_string_output_string(PORT *p)
{
  return(p->descr.s.data);
}

static void port_string_close(PORT *p)
{
  if (p->descr.s.data) 	scm_free(p->descr.s.data);
  scm_free(p);
}

static PORT *port_string_open(void *p, int flag)
{
  PORT *port = port_new(PORT_T_STRING, flag);

  if (flag & PORT_R) {
	port->descr.s.data = scm_must_strdup(p);
	port->descr.s.index= 0;
	port->descr.s.length = strlen(p);
	port->descr.s.alloced= port->descr.s.length+1;
  }
  return(port);
}

/************************************************************************
 * Port table driver
 ************************************************************************/

PORT_DESCR port_driver[PORT_MAX_TYPES] =
{
  { port_file_open, port_file_close,
	port_file_getc,	port_file_peekc, port_file_putc },

  { port_string_open, port_string_close,
	port_string_getc, port_string_peekc, port_string_putc }
};


/************************************************************************
 * Public interface
 ************************************************************************/

PORT *port_open(int type, void *p, int mode)
{
  return( (*port_driver[type].open)(p, mode) );
}

void port_close(PORT *p)
{
  (*port_driver[p->type].close)(p);
}

int port_getc(PORT *p)
{
  return( (*port_driver[p->type].getc)(p) );
}

int port_peekc(PORT *p)
{
  return( (*port_driver[p->type].peekc)(p) );
}

void port_putc(PORT *p, char c)
{
  (*port_driver[p->type].putc)(p, c);
}

/*-- high level interface */
PORT *port_open_input_file(char *fname)
{
  return(port_open(PORT_T_FILE, fname, PORT_R));
}

PORT *port_open_output_file(char *fname)
{
  return(port_open(PORT_T_FILE, fname, PORT_W));
}

PORT *port_open_input_string(char *string)
{
  return(port_open(PORT_T_STRING, string, PORT_R));
}

PORT *port_open_output_string()
{
  return(port_open(PORT_T_STRING, NULL, PORT_W));
}

void port_puts(PORT *p, char *str)
{
  while(*str)	port_putc(p, *str++);
}

void port_putn(PORT *p, long n)
{
  char buf[32];
  sprintf(buf, "%ld", n);
  port_puts(p, buf);
}

void port_putx(PORT *p, void *ptr)
{
  char buf[32];
  sprintf(buf, "%p", ptr);
  port_puts(p, buf);
}

void port_putd(PORT *p, double n)
{
  char buf[32];
  sprintf(buf, "%f", n);
  port_puts(p, buf);
}

/************************************************************************
 * Type related functions
 ************************************************************************/
SOBJ scm_mkport(PORT *port)
{
  SOBJ new = scm_newcell(SOBJ_T_PORT);
  SCM_PORT(new) = port;
  return(new);
}

void scm_port_sweep(SOBJ port)
{
  scm_puts("; gc: sweep port: "); scm_cprint(port);
  if (SCM_PORT(port) != NULL) port_close(SCM_PORT(port));
}

void scm_port_print(SOBJ port, PORT *p)
{
  port_puts(p, "<#port ");
  if (SCM_PORT(port)) {
	port_putx(p, SCM_PORT(port));
  } else {
	port_puts(p, "nil");
  }
  port_putc(p, '>');
}

void scm_eof_print(SOBJ port, PORT *p)
{
  port_puts(p, "#eof");
}

void scm_eof_write(SOBJ port, PORT *p)
{
  port_puts(p, "#eof");
}

/************************************************************************
 * scheme functions
 ************************************************************************/

void scm_write_obj(SOBJ obj, PORT *port, int raw)
{
  SOBJ p;

  if (port == NULL) port = scm_outp;

  if (obj == NULL) {	port_puts(port, "()"); 				return; }
  if (SCM_INUMP(obj)) { port_putn(port, SCM_INUM(obj)); 	return; }

  if (scm_is_opcode_address(obj)) {
	port_puts(port, "<#opc ");  port_puts(port, scm_search_opcode_name(obj));
	port_putc(port, '>');
	return;
  }
  if (obj >= (SOBJ)scm_stack && obj <= (SOBJ)scm_stack_limit) {
	port_puts(port, "<#sref "); port_putx(port, obj);  port_putc(port, '>');
	return;
  }
  if (!scm_is_pointer_to_heap(obj) && !scm_is_pointer_to_chr_array(obj)) {
	port_puts(port, "<#ptr ");  port_putx(port, obj);  port_putc(port, '>');
	return;
  }
  
  switch(obj->type & ~(SCM_GCMARK_MASK)) {
  case SOBJ_T_VOID:		port_puts(port, "#void");			break;
  case SOBJ_T_FNUM:		port_putd(port, SCM_FNUM(obj));		break;
  case SOBJ_T_BNUM:
	{
	  char *s = mpz_get_str(NULL, 10, SCM_BNUM(obj));
	  port_puts(port, s);
	  scm_free(s);
	  break;
	}
  case SOBJ_T_KEYWORD:
  case SOBJ_T_SYMBOL:
 	port_puts(port, SCM_ATOM_NAME(SCM_SYM_NAME(obj)));
	break;

  case SOBJ_T_LSYMBOL:
	port_puts(port, "<#lsym ");
	port_puts(port, SCM_ATOM_NAME(SCM_LSYM_NAME(obj)));
	port_putc(port, ' ');
	port_putn(port, SCM_LSYM_OFS(obj));
	port_putc(port, '>');
	break;

  case SOBJ_T_PAIR:
	port_putc(port, '(');
	for (p = obj; p; p = SCM_CDR(p)) {
	  if (p != obj) 	port_putc(port, ' ');
	  if (SCM_OBJTYPE(p) != SOBJ_T_PAIR) {
		if (SCM_OBJREF(p) != SCM_OBJREF(obj)) port_puts(port, ". ");
		scm_write_obj(p, port, raw);
		break;
	  }
	  scm_write_obj(SCM_CAR(p), port, raw);
	}
	port_putc(port, ')');
	break;
  case SOBJ_T_PRIM:
	port_puts(port, "<#prim ");	port_puts(port, SCM_PRIM(obj)->name);
	port_putc(port, ' ');			port_putn(port, SCM_PRIM(obj)->nargs);
	port_putc(port, '>');
	break;
  
  case SOBJ_T_CPRIM:
	port_puts(port, "<#cprim ");	port_putx(port, SCM_CPRIM_FUNC(obj));
	port_putc(port, ' ');			port_putn(port, SCM_CPRIM_NARGS(obj));
	port_putc(port, '>');
	break;
	
  case SOBJ_T_BOOLEAN:
	port_puts(port, (obj == scm_false) ? "#f" : "#t");
	break;
  case SOBJ_T_UNBOUND:		port_puts(port, "#unbound");  	break;
  case SOBJ_T_UNDEFINED:	port_puts(port, "#undefined");  break;
  case SOBJ_T_FREE:			port_puts(port, "<#free>");		break;
  default:
	if (obj->type < SOBJ_T_MAX) {
	  if (!raw && scm_type_hook[obj->type].write != NULL) {
		(*scm_type_hook[obj->type].write)(obj, port);
		return;
	  }
	  if (scm_type_hook[obj->type].print != NULL) {
		(*scm_type_hook[obj->type].print)(obj, port);
		return;
	  }
	}
	scm_puts("<#");	scm_puts(scm_type_hook[obj->type].name);
	scm_puts(" "); scm_putx(obj);
	scm_puts(">");
#ifdef COMMENT
	port_puts(scm_errp, "scm_write_obj: object type ");
	port_putn(scm_errp, obj->type);
	port_puts(scm_errp, " not reconized\n");
#endif
  }
}

/* get PORT from SCM_PORT : signal error if port is not a valid port*/
static PORT *get_port(SOBJ port)
{
  if (port == NULL) return(scm_outp);
  if (!SCM_PORTP(port)) err("bad port", port);
  return(SCM_PORT(port));
}

SOBJ scm_display2(SOBJ obj, SOBJ port)
{
  scm_write_obj(obj, get_port(port), 1);
  return(scm_undefined);
}

SOBJ scm_write2(SOBJ obj, SOBJ port)
{
  scm_write_obj(obj, get_port(port), 0);
  return(scm_undefined);
}

SOBJ scm_newline1(SOBJ port)
{
  port_puts(get_port(port), "\n");
  return(scm_undefined);
}

SOBJ scm_print2(SOBJ obj, SOBJ port)
{
  scm_display2(obj, port);
  scm_newline1(port);
  return(scm_undefined);
}

/*-- output to current-output port */
SOBJ scm_putc(int c)
{
  port_putc(scm_outp, c);  return(NULL);
}

SOBJ scm_puts(char *s)
{
  port_puts(scm_outp, s);  return(NULL);
}

SOBJ scm_putn(int n)
{
  port_putn(scm_outp, n);	return(NULL);
}

SOBJ scm_putx(void *ptr)
{
  port_putx(scm_outp, ptr);	return(NULL);
}

SOBJ scm_cdisplay(SOBJ obj)
{
  return(scm_display2(obj, NULL));
}

SOBJ scm_cprint(SOBJ obj)
{
  return(scm_print2(obj, NULL));
}

/*S* (call-with-input-file STRING PROC) */
/*S* (call-with-output-file STRING PROC) */
/*E* (call-with-input-string STRING PROC) */
/*E* (call-with-output-string PROC) */


/*E* (port? OBJ) */
SOBJ scm_portp(SOBJ x)
{
  return(SCM_MKBOOL(SCM_PORTP(x)));
}

/*S* (input-port? OBJ) */
SOBJ scm_input_portp(SOBJ x)
{
  return(SCM_MKBOOL( (SCM_PORTP(x) && 
					  SCM_PORT(x) &&
					  SCM_PORT(x)->type == PORT_R) ));
}

/*S* (output-port? OBJ) */
SOBJ scm_output_portp(SOBJ x)
{
  return(SCM_MKBOOL( (SCM_PORTP(x) && 
					  SCM_PORT(x) &&
					  SCM_PORT(x)->type == PORT_W) ));
}

/*S* (current-input-port) */
SOBJ scm_current_input_port()
{
  return(scm_in_port);
}

/*S* (current-output-port) */
SOBJ scm_current_output_port()
{
  return(scm_out_port);
}

/*E* (current-error-port) */
SOBJ scm_current_error_port()
{
  return(scm_err_port);
}

/*S* (with-input-from-file STRING THUNK) */
/*S* (with-output-to-file STRING THUNK) */

/*E* (with-input-from-string STRING THUNK) */
/*E* (with-output-to-string THUNK) */

/*S* (open-input-file STRING) */
SOBJ scm_open_input_file(SOBJ filename)
{
  SOBJ port;
  if (!SCM_STRINGP(filename)) err("open-input-file: bad filename", filename);
  port = scm_mkport(port_open_input_file(SCM_STR_VALUE(filename)));
  if (port == NULL) err("open-input-file: cannot open file", filename);
  return(port);
}

/*S* (open-output-file STRING) */
SOBJ scm_open_output_file(SOBJ filename)
{
  SOBJ port;
  if (!SCM_STRINGP(filename)) err("open-output-file: bad filename", filename);
  port = scm_mkport(port_open_output_file(SCM_STR_VALUE(filename)));
  if (port == NULL) err("open-output-file: cannot open file", filename);
  return(port);
}

/*E* (open-input-string STRING) */
/*E* (open-output-string) */
/*E* (get-output-string PORT) */

/*S* (close-port PORT) */
SOBJ scm_close_port(SOBJ port)
{
  if (!SCM_PORTP(port)) 	err("close-port: bad port", port);
  if (SCM_PORT(port))		port_close(SCM_PORT(port));
  SCM_PORT(port) = NULL;
  return(scm_true);
}

/*S* (close-input-port PORT) */
/*S* (close-output-port PORT) */

/*S* (read [PORT]) */
SOBJ scm_read(int argc, SOBJ *arg)
{
  PORT *port;
  
  if (argc >= 1) {
	if (!SCM_PORTP(arg[0])) 	err("read: bad port", arg[0]);
	port = SCM_PORT(arg[0]);
  } else {
	port = scm_inp;
  }
  return(scm_read_port(port));
}

/*S* (read-char [PORT]) */
SOBJ scm_read_char(int argc, SOBJ *arg)
{
  PORT *port;
  int c;

  if (argc >= 1) {
	if (!SCM_PORTP(arg[0])) 	err("read-char: bad port", arg[0]);
	port = SCM_PORT(arg[0]);
  } else {
	port = scm_inp;
  }
  if (port == NULL) 			err("read-char: port closed", NULL);
  if ((c = port_getc(port)) == PORT_EOF)
	return(scm_eof);
  return(scm_mkchar(c));
}

/*S* (peek-char [PORT]) */
SOBJ scm_peek_char(int argc, SOBJ *arg)
{
  PORT *port;
  int c;

  if (argc >= 1) {
	if (!SCM_PORTP(arg[0])) 	err("read-char: bad port", arg[0]);
	port = SCM_PORT(arg[0]);
  } else {
	port = scm_inp;
  }
  if (port == NULL) 			err("read-char: port closed", NULL);
  if ((c = port_peekc(port)) == PORT_EOF)
	return(scm_eof);
  return(scm_mkchar(c));
}

/*S* (eof-object? obj) */
SOBJ scm_eof_objectp(SOBJ obj)
{
  return(SCM_MKBOOL(obj == scm_eof));
}

/*S* (read-line PORT) */
SOBJ scm_read_line(SOBJ port)
{
  SOBJ str;
  int c;

  if (!SCM_PORTP(port)) err("read-line: bad port", port);

  str = scm_mkstring("");
  while( (c = port_getc(SCM_PORT(port))) != PORT_EOF) {
	if (c == '\n')
	  return(str);
	scm_string_append_char(str, SCM_MKINUM(c));
  }
  return(scm_eof);
}

/*S* (char-ready? [PORT]) */
SOBJ scm_char_readyp(int argc, SOBJ *arg)
{
  err("char-ready?: not implemented", NULL);
  return(NULL);
}

/*S* (write OBJ [PORT]) */
/*S* (display OBJ [PORT]) */
/*S* (newline [PORT]) */
/*S* (write-char CHAR [PORT]) */

SOBJ scm_write_char(int argc, SOBJ *arg)
{
  PORT *p;

  if (argc != 1 || argc != 2) 	err("write-char: bad number of args", NULL);

  if (!SCM_CHARP(arg[0])) err("write-char: bad char", arg[0]);
  if (argc == 1) {
	p = scm_outp;
  }	else {
	if (!SCM_PORTP(arg[1]))	err("write-char: bad port", arg[1]);
	p = SCM_PORT(arg[1]);
  }
  port_putc(p, SCM_CHAR(arg[0]));
  return(scm_true);
}

/*S* (load STRING) */
/*S* (transcript-on STRING) */
/*S* (transcript-off) */

/* The idea of this extensions is to bind FILE and PORT
 * Examples:
 * (define p (open-file "/tmp/tst.txt" "w+"))
 * (fseek (port-file p) 10 0)
 * (fputs (prot-file p) "Hello")
 * (close-port p)
 *
 * (let ((f (make-file-port (fopen "/tmp/tst.txt" "w"))))
 *    (fputs (port->file f) "hello world\n")
 *	  (port-close f))
 */

/*E* (open-file FILENAME MODE) -> PORT */
/*E* (open-string STRING MODE) -> PORT */
/*E* (make-file-port FILE) -> PORT */
/*E* (port->file PORT) -> FILE */


/************************************************************************
 * Initialization
 ************************************************************************/

void scm_port_init_default_files()
{

  /* reopen  stdin out and err to scm_inp, scm_outp, scm_errp ports*/
#ifdef KILL_STDIO
  int fn;

  fn = dup(0); fclose(stdin);   if (dup(fn) != 0) exit(2);
  scm_inp = port_new(PORT_T_FILE, PORT_R);
  scm_inp->descr.f = fdopen(0, "r");

  fn = dup(1); fclose(stdout);  if (dup(fn) != 1) exit(2);
  scm_outp = port_new(PORT_T_FILE, PORT_W);
  scm_outp->descr.f = fdopen(1, "w");
  
  fn = dup(2); fclose(stderr);  if (dup(fn) != 2) exit(2);
  scm_errp = port_new(PORT_T_FILE, PORT_W);
  scm_errp->descr.f = fdopen(2, "w");
#else
  scm_inp  = port_new(PORT_T_FILE, PORT_R);  scm_inp->descr.f = stdin;
  scm_outp = port_new(PORT_T_FILE, PORT_W);  scm_outp->descr.f = stdout;
  scm_errp = port_new(PORT_T_FILE, PORT_W);  scm_errp->descr.f = stderr;
#endif

}

void scm_init_port()
{
  scm_in_port = scm_mkport(scm_inp);  	scm_gc_protect(&scm_in_port);
  scm_out_port = scm_mkport(scm_outp);	scm_gc_protect(&scm_out_port);
  scm_err_port = scm_mkport(scm_errp);	scm_gc_protect(&scm_err_port);

  scm_eof= scm_newcell(SOBJ_T_EOF);		scm_gc_protect(&scm_eof);

  scm_add_cprim("port?",				scm_portp,					1);
  scm_add_cprim("input-port?",			scm_input_portp,			1);
  scm_add_cprim("output-port?",			scm_output_portp,			1);
  scm_add_cprim("current-input-port", 	scm_current_input_port, 	0);
  scm_add_cprim("current-output-port", 	scm_current_output_port, 	0);
  scm_add_cprim("current-error-port", 	scm_current_error_port, 	0);
  scm_add_cprim("open-input-file",		scm_open_input_file,		1);
  scm_add_cprim("open-output-file",		scm_open_output_file,		1);
  scm_add_cprim("close-port",			scm_close_port,				1);
  scm_add_cprim("read",					scm_read,					-1);
  scm_add_cprim("read-char",			scm_read_char,				-1);
  scm_add_cprim("peek-char",			scm_peek_char,				-1);
  scm_add_cprim("eof-object?",			scm_eof_objectp,			1);
  scm_add_cprim("char-ready?",			scm_char_readyp,			1);
  scm_add_cprim("read-line",			scm_read_line,				1);
  scm_add_cprim("write-char",			scm_write_char,				-1);
}
